''*****************************
''*  Hydra Asset Manager      *
''*  (C) Nurve Networks LLC.  *
''*****************************

CON

  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000 + 0000
  _stack = ($2400 + $2400 + 100) >> 2 'accomodate display memory and stack
  
var  long mypagenum 
     long mysize    
CON  mynumpages = 32
     mypagesize = 32

  x_tiles = 16
  y_tiles = 8
  screen_width = x_tiles * 16
  screen_height = y_tiles * 16
  
  paramcount = 14       
  front_buffer = $3800
  back_buffer  = $5C00

  CHAR_ROM_UPPER_IMAGE = $8000
  CHAR_ROM_LOWER_IMAGE = $8040
  CHAR_INDEX_01        = 24

  SERIAL_QUEUE_SIZE  = 2048
  SERIAL_BUFFER_SIZE = 1024
  EEPROM_BUFFER_SIZE = 1024

  MODE_WAITING        = 0           
  MODE_FILE_ADDRESS   = 1
  MODE_FILE_SIZE      = 2
  MODE_FILE_BUFFER    = 3
  MODE_FILE_BUFFER_OK = 4
  MODE_FILE_CRC       = 5
  MODE_EEPROM_SEND    = 10
  MODE_ERROR_SEND          = 100
  MODE_ERROR_READ_TOO_MUCH = 101
  
  TITLE_X = 128
  TITLE_Y = screen_height - 16 ' 8 * 16 - 16 = 112         
  
  STATUS_HEADER_X = 32
  STATUS_HEADER_Y = 96

  STATUS_X = 128
  STATUS_Y = 96

  FILENAME_HEADER_X = 32
  FILENAME_HEADER_Y = 80

  FILENAME_X = 96
  FILENAME_Y = 80

  BUFFER_X = 128
  BUFFER_Y = 20

  SECONDARY_STATUS_X = 80
  SECONDARY_STATUS_Y = 64

  STATUS_DOTS_X = 8
  STATUS_DOTS_Y = 16

  EEPROM_TEST_X = 128
  EEPROM_TEST_Y = 32

VAR

  long  tv_status     '0/1/2 = off/visible/invisible           read-only
  long  tv_enable     '0/? = off/on                            write-only
  long  tv_pins       '%ppmmm = pins                           write-only
  long  tv_mode       '%ccinp = chroma,interlace,ntsc/pal,swap write-only
  long  tv_screen     'pointer to screen (words)               write-only
  long  tv_colors     'pointer to colors (longs)               write-only               
  long  tv_hc         'horizontal cells                        write-only
  long  tv_vc         'vertical cells                          write-only
  long  tv_hx         'horizontal cell expansion               write-only
  long  tv_vx         'vertical cell expansion                 write-only
  long  tv_ho         'horizontal offset                       write-only
  long  tv_vo         'vertical offset                         write-only
  long  tv_broadcast  'broadcast frequency (Hz)                write-only
  long  tv_auralcog   'aural fm cog                            write-only

  word  screen[x_tiles * y_tiles]
  long  colors[64]

  long  serial_rx
  long  serial_rxidx
  long  serial_tx
  long  serial_txidx
  long  serial_rxbuf
  long  serial_rxsize
  long  serial_txbuf
  long  serial_txsize

  byte serial_rx_queue[2048]
  byte serial_rx_buffer[1024]
  byte eeprom_rx_buffer[EEPROM_BUFFER_SIZE]
  
  long  rx_size

  byte  external_string[32]
  byte  temp_serial_rx[32]

  byte sbuffer[32]

  long filesize
  long bufferSize
  long eepromOffset
  long bufferCRC

  long eepromWriteAddress
  long eepromWriteData

  byte updateCount
    
OBJ

  tv        : "tv_drv_010.spin"           'TV Driver  
  graphics  : "graphics_drv_010.spin"     'Graphics Driver
  eeprom    : "NS_eeprom_drv_011.spin"    'EEPROM Driver
  serial    : "copio_drv_001.spin"        'Serial COM Driver
 
PUB start      | i, j, k, kk, dx, dy, pp, pq, rr, numx, numchr

  'start tv
  longmove(@tv_status, @tvparams, paramcount)
  tv_screen := @screen
  tv_colors := @colors
  tv.start(@tv_status)

  'init colors
  repeat i from 0 to 64
    colors[i] := $00001010 * (i+4) & $F + $2B060C02

 'init tile screen
  repeat dx from 0 to tv_hc - 1
    repeat dy from 0 to tv_vc - 1
      screen[dy * tv_hc + dx] := front_buffer >> 6 + dy + dx * tv_vc + ((dy & $3F) << 10)

  'start and setup graphics
  graphics.start
  graphics.setup(x_tiles, y_tiles, 0, 0, back_buffer)

  ' start serial engine                                  
  serial.start(@serial_rx, 31, 30, 115200)
  serial.receive_circular(@serial_rx_queue, SERIAL_QUEUE_SIZE) ' Setup circular buffer for receive queue.

  'start eeprom driver
  eeprom.start(28, 29, 0)

  'Test reading from EEPROM.  Read 32 bytes at $8000, this is the first address past the 32k application.  
  eeprom.Read(@external_string, $8000, 32)
  repeat until eeprom.IsDone

  repeat

      'clear bitmap
      graphics.clear
      graphics.colorwidth(7,14)
      ' bits 15..31 are for square pixels
      ' bits  0..15 are for round pixels
      ' actual width is w[3..0]+1
      graphics.width($4)
      graphics.box(0, screen_height-16, screen_width, 16)
      graphics.textmode(1,1,6,4) ' 4 is horizontal center
      graphics.colorwidth(2,0)
      graphics.text(TITLE_X, TITLE_Y, @hamString)

      ' Simple animation to show the user that the app is still running.
      updateCount += 6
           
      if (updateCount =< 85)
         graphics.text(STATUS_DOTS_X, STATUS_DOTS_Y, string(".  "))
      elseif (updateCount =< 170)
         graphics.text(STATUS_DOTS_X, STATUS_DOTS_Y, string(" . "))
      else
         graphics.text(STATUS_DOTS_X, STATUS_DOTS_Y, string("  ."))

      ' check to make sure that rx_size isn't greater than SERIAL_BUFFER_SIZE
      if (rx_size > SERIAL_BUFFER_SIZE)
         driverMode := MODE_ERROR_READ_TOO_MUCH  

      ' -------------------------------------
      '            Waiting
      ' -------------------------------------      
      if (driverMode == MODE_WAITING)
         graphics.text(STATUS_X, STATUS_Y, string("Waiting"))
         ' Display the contents of the earlier read at $8000 to verify the data is correct.
         graphics.text(EEPROM_TEST_X, EEPROM_TEST_Y, @external_string)

         rx_size := serial.receive_queue(-1,-1)
         if (rx_size => 4)
            serial.receive_queue(@serial_rx_buffer, 4)
                     
            if (serial_rx_buffer[0] == fileString[0])
               serial.send(@fileString, 4)

               ' Now we have to wait for the server to validate the data.
               repeat
                  rx_size := serial.receive_queue(-1,-1)

                  if (rx_size => 2)
                     serial.receive_queue(@serial_rx_buffer, 2)

                     if (serial_rx_buffer[0] == okString[0])
                        driverMode := MODE_FILE_ADDRESS
                     quit                        
                     
            elseif ( (serial_rx_buffer[0] == eepromDownloadString[0]) )
               ' Init starting EEPROM read offset before we start reading.
               eepromOffset := 0
               eepromSize := 1 << (5 + serial_rx_buffer[3]) * 1024
               driverMode := MODE_EEPROM_SEND

      ' -------------------------------------
      '          Receive Address
      ' -------------------------------------      
      elseif (driverMode == MODE_FILE_ADDRESS)
         graphics.text(STATUS_X, STATUS_Y, string("Receiving Address"))

         rx_size := serial.receive_queue(-1,-1)
         if (rx_size => 4)
            serial.receive_queue(@serial_rx_buffer, 4)

            eepromAddress := long[@serial_rx_buffer]                        
            serial.send(@eepromAddress, 4)

            ' Now we have to wait for the server to validate the data.
            repeat
               rx_size := serial.receive_queue(-1,-1)

               if (rx_size => 2)
                  serial.receive_queue(@serial_rx_buffer, 2)

                  if (serial_rx_buffer[0] == okString[0])
                     driverMode := MODE_FILE_SIZE                     
                  quit                        

      ' -------------------------------------
      '         Receive Asset Size
      ' -------------------------------------      
      elseif (driverMode == MODE_FILE_SIZE)
         graphics.text(STATUS_X, STATUS_Y, string("Receiving Filesize"))

         rx_size := serial.receive_queue(-1,-1)
         if (rx_size => 4)

            serial.receive_queue(@serial_rx_buffer, 4)

            filesize := long[@serial_rx_buffer]

            serial.send(@filesize, 4)
            
            ' Now we have to wait for the server to validate the data.
            repeat
               rx_size := serial.receive_queue(-1,-1)

               if (rx_size => 2)
                  serial.receive_queue(@serial_rx_buffer, 2)

                  if (serial_rx_buffer[0] == okString[0])
                     driverMode := MODE_FILE_BUFFER
                  quit                        

      ' -------------------------------------
      '            Receive Buffer       
      ' -------------------------------------      
      elseif (driverMode == MODE_FILE_BUFFER)

            graphics.text(STATUS_X, STATUS_Y, string("Receiving Buffer"))
           
            bufferSize := ReadBufferToEEPROM(eepromAddress)
            ' Offset eeprom address so the next buffer's worth goes into the correct location.
            eepromAddress += bufferSize
            filesize -= bufferSize
           
            UpdateDisplayBuffer
            graphics.text(BUFFER_X, BUFFER_Y, @temp_serial_rx)

         if (filesize == 0)           
            driverMode := MODE_WAITING

      ' -------------------------------------
      '    Send Buffer's Worth of EEPROM       
      ' -------------------------------------      
      elseif (driverMode == MODE_EEPROM_SEND)
         graphics.text(STATUS_X, STATUS_Y, string("Sending EEPROM Contents to PC"))

         repeat mypagenum from 0 to mynumpages-1
           eeprom.Read(@eeprom_rx_buffer+(mypagenum*mypagesize), eepromOffset+(mypagenum*mypagesize), mypagesize)
           repeat until eeprom.IsDone

         eepromOffset += 1024   
    
         ' Now, send what we received from the eeprom read back to the PC   
         if ( serial.send(@eeprom_rx_buffer, 1024) == false )
            driverMode := MODE_ERROR_SEND
         else
            ' Now we have to wait for the server to validate the data.
            rx_size := 0
            repeat until (rx_size => 2)
               rx_size := serial.receive_queue(-1,-1)

            serial.receive_queue(@serial_rx_buffer, 2)

            ' Currently, we're not checking the returned string from the PC.   As long as it sends 2 bytes, we continue.

            ' See if we have sent all the EEPROM buffers. 
            if (eepromOffset == eepromSize)
               driverMode := MODE_WAITING      
         
      elseif (driverMode == MODE_ERROR_SEND)
         graphics.text(STATUS_X, STATUS_Y, string("Send Error!"))

      elseif (driverMode == MODE_ERROR_READ_TOO_MUCH)
         graphics.text(STATUS_X, STATUS_Y, string("Read Buffer Overflow!"))
         
      graphics.copy(front_buffer)
      
  'end repeat - main loop

PUB ClearMemory | i

   ' Clears eepromSize with a fill value of 0.
   
   ' Setup a buffer with nothing but zeros.
   repeat i from 0 to 511
      eeprom_rx_buffer[i] := 0

   ' Keep writing the zeroed out buffer to eeprom memory   
   repeat i from 0 to (eepromSize/mypagesize)

      repeat mypagenum from 0 to mynumpages-1
        eeprom.Read(@eeprom_rx_buffer+(mypagenum*mypagesize), (i*mypagesize)+(mypagenum*mypagesize), mypagesize)
        repeat until eeprom.IsDone

'      eeprom.Write(@eeprom_rx_buffer, i*512, 512)
'      repeat until eeprom.IsDone
   
PUB MemoryWriteTest | i

   ' Writes a unique value in each of the 32768 longs in eepromSize.
   
   repeat i from 0 to (eepromSize/4)
      eepromWriteData := i*4

      ' Write a 1 into our test location.   
      eeprom.Write(@eepromWriteData, i*4, 4)
      repeat until eeprom.IsDone


PUB ReadBufferToEEPROM(eeprom_address) : buffer_size | i

   ' Wait until we have 4 bytes.  It's the size of the buffer we're going to read in.
   rx_size := 0
   repeat until (rx_size => 4)
      rx_size := serial.receive_queue(-1,-1)
                  
   ' We received some data from the server.
   serial.receive_queue(@serial_rx_buffer, 4)
   buffer_size := long[@serial_rx_buffer]

   rx_size := 0
   repeat until (rx_size => buffer_size )
      rx_size := serial.receive_queue(-1,-1)

   serial.receive_queue(@serial_rx_buffer, buffer_size)

   ' Write the data to the eeprom address that we received in the previous step.
   'rb - TODO - We need to check to make sure that bufferSize is not greater than EEPROM write buffer.
   mypagenum:=0
   mysize:=buffer_size
   repeat while mysize>mypagesize
     eeprom.Write(@serial_rx_buffer+(mypagenum*mypagesize), eeprom_address+(mypagenum*mypagesize), mypagesize)
     repeat until eeprom.IsDone
     mypagenum++
     mysize-=mypagesize
   if mysize>0
     eeprom.Write(@serial_rx_buffer+(mypagenum*mypagesize), eeprom_address+(mypagenum*mypagesize), mysize)
     repeat until eeprom.IsDone
'   eeprom.Write(@serial_rx_buffer, eeprom_address, buffer_size)
'   repeat until eeprom.IsDone

   ' This is currently garbage now that we're not checking CRC.  For now, just send two bytes as a synchronization method between
   ' the Hydra and PC.   Later we'll want the user to be able to enable CRC checks.
   serial.send(@bufferCRC, 2)

   return buffer_size      

PUB UpdateDisplayBuffer | i

  ' Copy 32 bytes of the serial_rx_buffer into a temp buffer for visual display of the incoming stream.
  repeat i from 0 to 31
    if ((serial_rx_buffer[i+2] > 32) and (serial_rx_buffer[i+2] < 128))
      temp_serial_rx[i] := serial_rx_buffer[i+2]
    else
      temp_serial_rx[i] := $2E  ' Use a period if it's an extended character.
      
    temp_serial_rx[31] := 0

PUB itoa(value, string_ptr)     | base, factor, i
' converts an integer into an ASCIIZ, up to 99_999x
' string_ptr points to target storage to store converted string

' 100,000's place
base := 100_000

repeat i from 6 to 2
  factor := value / base 
  byte[0][string_ptr++] := hex_table [ factor ]
  value -= factor*base
  base /= 10

' 1's position
byte[0][string_ptr++] := hex_table [ value ]

' NULL terminate
byte[0][string_ptr++] := 0
  
DAT

tvparams                long    0               'status
                        long    1               'enable
'                        long    %011_0001       'pins - wireless video
                        long    %011_0000       'pins - normal video 
                        long    %0000           'mode
                        long    0               'screen
                        long    0               'colors
                        long    x_tiles         'hc
                        long    y_tiles         'vc
                        long    10              'hx
                        long    1               'vx
                        long    0               'ho
                        long    0               'vo
                        long    60_000_000      '_xinfreq<<4  'broadcast
                        long    0               'auralcog

hamString               byte    "Hydra Asset Manager - 1.0.9",0
fileString              byte    "FILE",0                         
okString                byte    "OK"
driverMode              byte    MODE_WAITING
eepromAddress           long    0 
eepromSize                              long    128*1024
hex_table               byte    "0123456789ABCDEF"
eepromDownloadString    byte    "DLEP"
                                     